@@ -1,5 +1,7 @@
 //! Tools for text file manipulation
 
+// column specs need globs, +=1 and -=1, meaning maybe they have to be signed
+
 #![warn(
     missing_copy_implementations,
     absolute_paths_not_starting_with_crate,
@@ -26,12 +28,14 @@ use flate2::read::MultiGzDecoder;
 use fs_err as fs;
 use std::error;
 use std::io::{self, BufRead, Read, Write};
+use std::ops::{Deref, DerefMut};
 use std::{fmt, str};
 
 pub mod check;
 pub mod column;
 pub mod comp;
 pub mod expr;
+pub mod join;
 pub mod sort;
 pub mod tooltest;
 
@@ -51,6 +55,8 @@ pub enum Error {
     Error(String),
     /// pass through ParseIntError
     ParseIntError(std::num::ParseIntError),
+    /// pass through ParseIntError
+    ParseFloatError(std::num::ParseFloatError),
     /// pass through io::Error
     IoError(std::io::Error),
     /// pass through regex::Error
@@ -96,11 +102,18 @@ impl From<std::num::ParseIntError> for Error {
     }
 }
 
+impl From<std::num::ParseFloatError> for Error {
+    fn from(kind: std::num::ParseFloatError) -> Error {
+        Error::ParseFloatError(kind)
+    }
+}
+
 impl fmt::Display for Error {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
             Error::Error(s) => write!(f, "Cdx Error : {}", s)?,
             Error::ParseIntError(s) => write!(f, "ParseIntError : {}", s)?,
+            Error::ParseFloatError(s) => write!(f, "ParseFloatError : {}", s)?,
             Error::IoError(s) => write!(f, "IoError : {}", s)?,
             Error::RegexError(s) => write!(f, "RegexError : {}", s)?,
             Error::NeedLookup => write!(
@@ -114,7 +127,7 @@ impl fmt::Display for Error {
 }
 
 /// pointers into a vector, simulating a slice without the ownership issues
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, Clone, Copy, Default)]
 pub struct FakeSlice {
     begin: u32,
     end: u32,
@@ -136,9 +149,9 @@ pub struct FakeSlice {
 /// assert_eq!(line.len(), 3);
 /// assert_eq!(line.get(1), b"two");
 ///```
-#[derive(Debug, Clone)]
+#[derive(Debug, Clone, Default)]
 pub struct TextLine {
-    /// The whole input line, without newline
+    /// The whole input line, with newline
     pub line: Vec<u8>,
     /// the individual columns
     pub parts: Vec<FakeSlice>,
@@ -232,12 +245,6 @@ impl TextLine {
     }
 }
 
-impl Default for TextLine {
-    fn default() -> Self {
-        Self::new()
-    }
-}
-
 impl<'a> IntoIterator for &'a TextLine {
     type Item = &'a [u8];
     type IntoIter = TextLineIter<'a>;
@@ -277,7 +284,6 @@ pub struct Infile {
     pub f: io::BufReader<Box<dyn Read>>,
 }
 
-// Should deref to f
 impl Infile {
     /// create a new input file
     pub fn new(f: io::BufReader<Box<dyn Read>>) -> Self {
@@ -285,12 +291,32 @@ impl Infile {
     }
 }
 
+impl Default for Infile {
+    fn default() -> Self {
+        Self::new(io::BufReader::new(Box::new(io::empty())))
+    }
+}
+
 impl fmt::Debug for Infile {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(f, "Infile")
     }
 }
 
+impl Deref for Infile {
+    type Target = io::BufReader<Box<dyn Read>>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.f
+    }
+}
+
+impl DerefMut for Infile {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        &mut self.f
+    }
+}
+
 type Outfile = io::BufWriter<Box<dyn Write>>;
 
 /// Make an Outfile from a file name
@@ -326,7 +352,7 @@ pub fn get_reader(name: &str) -> Result<Infile> {
     Ok(Infile::new(outer))
 }
 
-#[derive(Debug)]
+#[derive(Debug, Default)]
 struct InfileContext {
     header: TextLine,
     delim: u8,
@@ -384,13 +410,7 @@ impl InfileContext {
     }
 }
 
-impl Default for InfileContext {
-    fn default() -> Self {
-        Self::new()
-    }
-}
-
-#[derive(Debug)]
+#[derive(Debug, Default)]
 /// File reader for text file broken into lines with columns
 pub struct Reader {
     file: Infile,
@@ -405,7 +425,7 @@ impl Reader {
     /// make a new Reader
     pub fn new() -> Self {
         Self {
-            file: Infile::new(io::BufReader::new(Box::new(io::empty()))),
+            file: Infile::default(),
             line: TextLine::new(),
             cont: InfileContext::new(),
             do_split: true,
@@ -454,13 +474,7 @@ impl Reader {
     }
 }
 
-impl Default for Reader {
-    fn default() -> Self {
-        Self::new()
-    }
-}
-
-#[derive(Debug)]
+#[derive(Debug, Default)]
 /// File reader for text file broken into lines with columns
 /// previous N lines are stil available
 pub struct LookbackReader {
@@ -478,7 +492,7 @@ impl LookbackReader {
         let mut lines: Vec<TextLine> = Vec::new();
         lines.resize(lookback + 1, TextLine::new());
         Self {
-            file: Infile::new(io::BufReader::new(Box::new(io::empty()))),
+            file: Infile::default(),
             lines,
             cont: InfileContext::new(),
             do_split: true,
@@ -547,12 +561,6 @@ impl LookbackReader {
     }
 }
 
-impl Default for LookbackReader {
-    fn default() -> Self {
-        Self::new(1)
-    }
-}
-
 fn prerr(data: &[&[u8]]) -> Result<()> {
     for x in data {
         std::io::stderr().write_all(x)?;
@@ -560,3 +568,213 @@ fn prerr(data: &[&[u8]]) -> Result<()> {
     std::io::stderr().write_all(b"\n")?;
     Ok(())
 }
+
+fn bytes_equal(c1: u8, c2: u8, ic: bool) -> bool {
+    if c1 == c2 {
+        return true;
+    }
+    if c1 == b'?' {
+        return true;
+    }
+    if !ic {
+        return false;
+    }
+    c1.to_ascii_lowercase() == c2.to_ascii_lowercase()
+}
+
+fn chars_equal(c1: char, c2: char, ic: bool) -> bool {
+    if c1 == c2 {
+        return true;
+    }
+    if c1 == '?' {
+        return true;
+    }
+    if !ic {
+        return false;
+    }
+    c1.to_lowercase().eq(c2.to_lowercase())
+}
+
+/// match in glob format
+pub fn bglob(mut wild: &[u8], mut buff: &[u8], ic: bool) -> bool {
+    while !buff.is_empty() && !wild.is_empty() && (wild[0] != b'*') {
+        if !bytes_equal(wild[0], buff[0], ic) {
+            return false;
+        }
+        wild = &wild[1..];
+        buff = &buff[1..];
+    }
+    if wild.is_empty() && !buff.is_empty() {
+        return false;
+    }
+
+    let mut cp: &[u8] = &[];
+    let mut mp: &[u8] = &[];
+
+    while !buff.is_empty() {
+        if !wild.is_empty() && (wild[0] == b'*') {
+            wild = &wild[1..];
+            if wild.is_empty() {
+                return true;
+            }
+            mp = wild;
+            cp = &buff[1..];
+        } else if !wild.is_empty() && bytes_equal(wild[0], buff[0], ic) {
+            wild = &wild[1..];
+            buff = &buff[1..];
+        } else {
+            wild = mp;
+            cp = &cp[1..];
+            buff = cp;
+        }
+    }
+
+    while !wild.is_empty() && (wild[0] == b'*') {
+        wild = &wild[1..];
+    }
+    wild.is_empty()
+}
+
+fn first(s: &str) -> char {
+    s.chars().next().unwrap()
+}
+
+//fn first_len(s : &str) -> usize {
+//    s.chars().next().unwrap().len_utf8()
+//}
+
+fn skip_first(s: &str) -> &str {
+    &s[s.chars().next().unwrap().len_utf8()..]
+}
+
+fn take_first(slice: &mut &str) -> char {
+    let x = slice.chars().next().unwrap();
+    *slice = &slice[x.len_utf8()..];
+    x
+}
+
+/// match in glob format
+pub fn sglob(mut wild: &str, mut buff: &str, ic: bool) -> bool {
+    while !buff.is_empty() && !wild.is_empty() && (first(wild) != '*') {
+        if !chars_equal(first(wild), first(buff), ic) {
+            return false;
+        }
+        wild = skip_first(wild);
+        buff = skip_first(buff);
+    }
+    if wild.is_empty() && !buff.is_empty() {
+        return false;
+    }
+
+    let mut cp: &str = "";
+    let mut mp: &str = "";
+
+    while !buff.is_empty() {
+        if !wild.is_empty() && (first(wild) == '*') {
+            wild = &wild[1..];
+            if wild.is_empty() {
+                return true;
+            }
+            mp = wild;
+            cp = skip_first(buff);
+        } else if !wild.is_empty() && chars_equal(first(wild), first(buff), ic) {
+            wild = skip_first(wild);
+            buff = skip_first(buff);
+        } else {
+            wild = mp;
+            cp = skip_first(cp);
+            buff = cp;
+        }
+    }
+
+    while !wild.is_empty() && (first(wild) == '*') {
+        wild = skip_first(wild);
+    }
+    wild.is_empty()
+}
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    #[test]
+    fn uwild() {
+        assert!(bglob(b"", b"", false));
+        assert!(!bglob(b"", b"foo", false));
+        assert!(!bglob(b"foo", b"", false));
+        assert!(bglob(b"*", b"foo", false));
+        assert!(bglob(b"*", b"", false));
+
+        assert!(bglob(b"foo", b"foo", false));
+        assert!(!bglob(b"bar", b"foo", false));
+        assert!(bglob(b"foo", b"FoO", true));
+        assert!(!bglob(b"bar", b"FoO", false));
+
+        assert!(!bglob(b"*.txt", b"foo.txtt", false));
+        assert!(bglob(b"*.txt", b"foo.txt", false));
+        assert!(bglob(b"foo*.txt", b"foo.txt", false));
+        assert!(!bglob(b"foo*.txt", b"foo.txtt", false));
+        assert!(bglob(b"foo?txt", b"foo.txt", false));
+        assert!(!bglob(b"foo?txt", b"foo..txt", false));
+
+        assert!(bglob(b"*.txt", b"foO.txt", true));
+        assert!(!bglob(b"*.txt", b"foO.txtt", true));
+        assert!(bglob(b"foo*.txt", b"foO.txt", true));
+        assert!(!bglob(b"foo*.txt", b"foO.txtt", true));
+        assert!(bglob(b"foo?txt", b"foO.txt", true));
+        assert!(!bglob(b"foo?txt", b"foO..txt", true));
+
+        assert!(bglob(b"a?b", b"anb", false));
+        assert!(!bglob(b"a?b", "añb".as_bytes(), false));
+
+        assert!(bglob(
+	    b"s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/*/add_hour_*",
+	    b"s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/2da/add_hour_201102122300_2da.gz", false));
+        assert!(bglob(
+	    b"s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/???/add_hour_*",
+	    b"s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/2da/add_hour_201102122300_2da.gz", false));
+    }
+    // ñ
+    #[test]
+    fn swild() {
+        assert!(sglob("", "", false));
+        assert!(!sglob("", "foo", false));
+        assert!(!sglob("foo", "", false));
+        assert!(sglob("*", "foo", false));
+        assert!(sglob("*", "", false));
+
+        assert!(sglob("foo", "foo", false));
+        assert!(!sglob("bar", "foo", false));
+        assert!(sglob("foo", "FoO", true));
+        assert!(!sglob("bar", "FoO", false));
+
+        assert!(!sglob("*.txt", "foo.txtt", false));
+        assert!(sglob("*.txt", "foo.txt", false));
+        assert!(sglob("foo*.txt", "foo.txt", false));
+        assert!(!sglob("foo*.txt", "foo.txtt", false));
+        assert!(sglob("foo?txt", "foo.txt", false));
+        assert!(!sglob("foo?txt", "foo..txt", false));
+
+        assert!(sglob("*.txt", "foO.txt", true));
+        assert!(!sglob("*.txt", "foO.txtt", true));
+        assert!(sglob("foo*.txt", "foO.txt", true));
+        assert!(!sglob("foo*.txt", "foO.txtt", true));
+        assert!(sglob("foo?txt", "foO.txt", true));
+        assert!(!sglob("foo?txt", "foO..txt", true));
+
+        assert!(sglob("a?b", "anb", false));
+        assert!(sglob("a?b", "añb", false));
+
+        assert!(!sglob("añb", "aÑb", false));
+        assert!(!sglob("aÑb", "añb", false));
+        assert!(sglob("añb", "aÑb", true));
+        assert!(sglob("aÑb", "añb", true));
+
+        assert!(sglob(
+	    "s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/*/add_hour_*",
+	    "s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/2da/add_hour_201102122300_2da.gz", false));
+        assert!(sglob(
+	    "s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/???/add_hour_*",
+	    "s3://com.company.metric-holding-bin-prod//2011/02/12/23/00/acctid/2da/add_hour_201102122300_2da.gz", false));
+    }
+}
